import itertools
import bisect


def sorted_chain(*ranges):
    return sorted(itertools.chain(*ranges))


def generate_range_checker(sorted_ranges):

    ranges_start = [t[0] for t in sorted_ranges]

    def _char_in_range(char):
        code_point = ord(char)

        # 1. find a range such that (start, end), start <= code_point.
        idx = bisect.bisect_left(ranges_start, code_point)
        if idx == len(ranges_start) or \
                (idx != 0 and code_point != ranges_start[idx]):
            idx -= 1

        # 2. check if start <= code_point <= end.
        # (to deal with the coner case when idx == 0).
        return sorted_ranges[idx][0] <= code_point <= sorted_ranges[idx][1]

    return _char_in_range


def fullwidth_to_halfwidth(seq):

    def convert(char):
        code_point = ord(char)
        if not (0xFF01 <= code_point <= 0xFF5E):
            return char
        return chr(code_point - 0xFEE0)

    return ''.join(map(convert, seq))


def _flatten_nested(seq, ret=None):
    if ret is None:
        ret = []
    for item in seq:
        if not isinstance(item, (list, tuple, set)):
            ret.append(item)
        else:
            _flatten_nested(item, ret)
    return ret


def _append_code_points_to_text(text, *code_points):
    return [
        text + chr(cp)
        for cp in code_points
    ]


def _append_code_points_to_seq(seq, *code_points):
    if isinstance(seq, str):
        seq = [seq]
    return _flatten_nested([
        _append_code_points_to_text(text, *code_points)
        for text in seq
    ])


def _single_quotation(seq):
    return _append_code_points_to_seq(seq, 0xFF07, 0x2019, 0x2032)


def _double_quotation(seq):
    return _append_code_points_to_seq(seq, 0xFF02, 0x201D, 0x2033)


SENTENCE_ENDS = _flatten_nested([
    # size 3.
    _double_quotation('？！'),
    _double_quotation('……'),
    _double_quotation(
        _single_quotation('。'),
    ),
    _double_quotation(
        _single_quotation('！'),
    ),

    # size 2.
    _double_quotation('。'),
    _double_quotation('！'),
    _double_quotation('？'),
    _double_quotation('；'),
    '？！',

    # size 1.
    '…',
    '。',
    '！',
    '？',
    '；',
])
# add normalized endings.
SENTENCE_ENDS = _flatten_nested([
    set((
        end,
        fullwidth_to_halfwidth(end),
    ))
    for end in SENTENCE_ENDS
])


# English Chars.
ENGLISH_CHARS = sorted_chain(
    # ASCII_ALPHA_RANGES
    [
        (0x0041, 0x005A),
        (0x0061, 0x007A),
    ],
    # ALPHA_EXTENSION_RANGES
    [
        (0xFF21, 0xFF3A),
        (0xFF41, 0xFF5A),
    ]
)


# Digits.
DIGITS = sorted_chain(
    # ASCII_DIGIT_RANGES
    [
        (0x0030, 0x0039),
    ],
    # DIGIT_EXTENSION_RANGES
    [
        (0xFF10, 0xFF19),
    ],
)


# Delimiters.
DELIMITERS = sorted_chain(
    # ASCII_DELIMITERS_RANGES
    [
        (0x0021, 0x002F),
        (0x003A, 0x0040),
        (0x005B, 0x0060),
        (0x007B, 0x007E),
    ],
    # GENERAL_DELIMITERS_RAGES
    # http://www.unicode.org/charts/PDF/U2000.pdf
    [
        (0x2000, 0x206F),
    ],
    # CJK_DELIMITERS_RANGES
    # http://www.unicode.org/charts/PDF/U3000.pdf
    # http://www.unicode.org/charts/PDF/UFE30.pdf
    [
        (0x3000, 0x303F),
        (0xFE30, 0xFE4F),
    ],
    # DELIMITERS_EXTENSION_RANGES
    # http://www.unicode.org/charts/PDF/UFF00.pdf
    [
        (0xFF01, 0xFF0F),
        (0xFF1A, 0xFF20),
        (0xFF3B, 0xFF40),
        (0xFF5B, 0xFF64),
        (0xFFE0, 0xFFEE),
    ],
)


# Chinese Chars.
# pull from https://www.qqxiuzi.cn/zh/hanzi-unicode-bianma.php
# notice 3007 a delimiter, hence should not be included.
#
# lines = '''copy paste'''
# [l.split('\t') for l in lines.strip().split('\n')]
#
CHINESE_CHARS = sorted_chain(
    [
        (0x4E00, 0x9FA5),
        (0x9FA6, 0x9FEF),
        (0x3400, 0x4DB5),
        (0x20000, 0x2A6D6),
        (0x2A700, 0x2B734),
        (0x2B740, 0x2B81D),
        (0x2B820, 0x2CEA1),
        (0x2CEB0, 0x2EBE0),
        (0x2F00, 0x2FD5),
        (0x2E80, 0x2EF3),
        (0xF900, 0xFAD9),
        (0x2F800, 0x2FA1D),
        (0xE815, 0xE86F),
        (0xE400, 0xE5E8),
        (0xE600, 0xE6CF),
        (0x31C0, 0x31E3),
        (0x2FF0, 0x2FFB),
        (0x3105, 0x312F),
        (0x31A0, 0x31BA),
    ],
)



from .const import (
    CHINESE_CHARS,
    ENGLISH_CHARS,
    DIGITS,
    sorted_chain,
)
from .common import (
    generate_ranges_marker,
    generate_segmenter,
)


SEGMENT_RANGES = sorted_chain(
    CHINESE_CHARS,
    ENGLISH_CHARS,
    DIGITS,
)


def dlmseg_start_cond_fn(start, segments):
    return segments[start]


def dlmseg_end_cond_fn(end, segments):
    if segments[end]:
        return False, end + 1
    else:
        return True, end


dlmseg = generate_segmenter(
    [
        generate_ranges_marker(SEGMENT_RANGES),
    ],
    dlmseg_start_cond_fn,
    dlmseg_end_cond_fn,
)



import re

from .const import (
    CHINESE_CHARS,
    ENGLISH_CHARS,
    DIGITS,
    DELIMITERS,
)


def generate_replacer(sorted_ranges):

    def ranges_to_pattern(sorted_ranges):
        inner = [
            rf'{chr(lb)}-{chr(ub)}'
            for lb, ub in sorted_ranges
        ]
        inner = ''.join(inner)
        return rf'[{inner}]+'

    pattern = re.compile(ranges_to_pattern(sorted_ranges), re.UNICODE)

    def replacer(text, repl=''):
        return pattern.sub(repl, text)

    return replacer


replace_chinese_chars = generate_replacer(CHINESE_CHARS)
replace_english_chars = generate_replacer(ENGLISH_CHARS)
replace_digits = generate_replacer(DIGITS)
replace_delimiters = generate_replacer(DELIMITERS)



from collections import Counter

from cnt.rulebase.sentseg import (
    mark_whitespaces,
    mark_extended_chinese_chars,
    mark_sentence_endings,
    sentseg,
    SENTSEG_RANGES,
    SENTENCE_ENDS,
)


def test_sentseg_ends():
    assert len(SENTENCE_ENDS) == len(set(SENTENCE_ENDS))


def test_sentseg():
    text = (
        '测试句子一， 测试。  测试 句子二。 测试句子三！！！'
    )
    sents = sentseg(text)

    assert 3 == len(sents)

    sent1_text, sent1_range = sents[0]
    assert '测试句子一， 测试。' == sent1_text
    assert '测试句子一， 测试。' == text[sent1_range[0]:sent1_range[1]]

    sent2_text, sent2_range = sents[1]
    assert '测试 句子二。' == sent2_text
    assert '测试 句子二。' == text[sent2_range[0]:sent2_range[1]]

    sent3_text, sent3_range = sents[2]
    assert '测试句子三！！！' == sent3_text
    assert '测试句子三！！！' == text[sent3_range[0]:sent3_range[1]]

    text = (
        'a,b，c‚d'
    )
    assert 4 == len(sentseg(text, enable_comma=True))


def test_no_overlapping_char_ranges():
    pre_end = -1
    for start, end in SENTSEG_RANGES:
        assert end >= start
        assert start > pre_end
        pre_end = end


def test_mark_whitespaces():
    text = 'aa  \n a\ta '
    assert [
        False, False,
        True, True, True, True,
        False,
        True,
        False,
        True,
    ] == mark_whitespaces(text)


def test_mark_extended_chinese_chars():
    text = '  测试。 \tabc '
    assert [
        False, False,
        True, True, True,
        False, False,
        True, True, True,
        False,
    ] == mark_extended_chinese_chars(text)


def test_mark_sentence_endings():
    text = 'a。"b，c!？！”d!！!'
    assert [
        False,
        True, True,
        False, False, False,
        True, True, True, True,
        False,
        True, True, True,
    ] == mark_sentence_endings(text)



from cnt.rulebase.replace import (
    replace_chinese_chars,
    replace_english_chars,
    replace_digits,
    replace_delimiters,
)


def test_replacers():
    assert '12' == replace_chinese_chars('1测试2')
    assert '1 2' == replace_chinese_chars('1测试2', ' ')

    assert '12' == replace_english_chars('1english2')
    assert 'abc' == replace_digits('a1b2c3')
    assert '12' == replace_delimiters('1,2,')



# -*- coding: utf-8 -*-

"""Top-level package for cnt.rulebase."""

__author__ = """Hunt Zhan"""
__email__ = 'huntzhan.dev@gmail.com'
__version__ = '0.5.0'


from .sentseg import sentseg  # noqa: F401
from .dlmseg import dlmseg  # noqa: F401
from .replace import (
    replace_chinese_chars,
    replace_english_chars,
    replace_digits,
    replace_delimiters,
)  # noqa: F401



from .const import generate_range_checker


def generate_ranges_marker(ranges):

    def ranges_marker(
        text,
        _ranges_checker=generate_range_checker(ranges),
    ):
        marks = [False] * len(text)
        for idx, c in enumerate(text):
            if _ranges_checker(c):
                marks[idx] = True
        return marks

    return ranges_marker


def generate_segmenter(markers, start_cond_fn, end_cond_fn):

    def segmenter(text):
        marks_group = [
            m(text) for m in markers
        ]

        # two pointers move.
        sentences = []

        def _push_to_sentence(start, end):
            sentences.append((
                text[start:end],
                (start, end),
            ))
            return end

        TEXTLEN = len(text)
        start, end = 0, 0
        while start < TEXTLEN:
            if not start_cond_fn(start, *marks_group):
                start += 1
                continue

            end = start + 1
            while end < TEXTLEN:
                should_break, end = end_cond_fn(end, *marks_group)
                if should_break:
                    break

            start = _push_to_sentence(start, end)

        if start < TEXTLEN:
            _push_to_sentence(start, TEXTLEN)

        return sentences

    return segmenter



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


def load_requirements(path):
    with open(path) as fin:
        return [
            line
            for line in map(lambda l: l.strip(), fin.readlines())
            if line and not line.startswith('#')
        ]


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = load_requirements('requirements_prod.txt')
test_requirements = load_requirements('requirements_dev.txt')

setup(
    author="Hunt Zhan",
    author_email='huntzhan.dev@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="None",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='cnt.rulebase',
    name='cnt.rulebase',
    packages=['cnt.rulebase'],
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/huntzhan/cnt.rulebase',
    version='0.5.0',
    zip_safe=False,
)



import re

from .const import (
    SENTENCE_ENDS,
    CHINESE_CHARS,
    ENGLISH_CHARS,
    DIGITS,
    DELIMITERS,
    sorted_chain,
)
from .common import (
    generate_ranges_marker,
    generate_segmenter,
)

import ahocorasick


SENTSEG_RANGES = sorted_chain(
    CHINESE_CHARS,
    ENGLISH_CHARS,
    DIGITS,
    DELIMITERS,
)


def mark_whitespaces(
    text,
    _whitespace_pattern=re.compile(r'\s+'),
):
    '''
    return a list of marks for identifying whitespaces.
    '''
    marks = [False] * len(text)
    for match in _whitespace_pattern.finditer(text):
        start, end = match.span()
        marks[start:end] = (True,) * (end - start)
    return marks


mark_extended_chinese_chars = generate_ranges_marker(SENTSEG_RANGES)


def _build_ac_automation(keys):
    ac = ahocorasick.Automaton()
    for idx, key in enumerate(keys):
        ac.add_word(key, (idx, key))
    ac.make_automaton()
    return ac


def mark_sentence_endings(
    text,
    _ac_automation=_build_ac_automation(SENTENCE_ENDS),
):
    marks = [False] * len(text)
    for end, (_, key) in _ac_automation.iter(text):
        end += 1
        start = end - len(key)
        marks[start:end] = (True,) * (end - start)
    return marks


def mark_sentence_endings_with_comma(
    text,
    _ac_automation=_build_ac_automation(
        SENTENCE_ENDS + [chr(0xFF0C), chr(0x201A), ','],
    ),
):
    return mark_sentence_endings(text, _ac_automation=_ac_automation)


def sentseg_start_cond_fn(
    start,
    whitespaces, extended_chinese_chars, sentence_endings,
):
    return extended_chinese_chars[start]


def sentseg_end_cond_fn(
    end,
    whitespaces, extended_chinese_chars, sentence_endings,
):
    if not (extended_chinese_chars[end] or whitespaces[end]):
        return True, end
    elif sentence_endings[end]:
        while end < len(sentence_endings) and sentence_endings[end]:
            end += 1
        return True, end
    else:
        return False, end + 1


_sentseg = generate_segmenter(
    [
        mark_whitespaces,
        mark_extended_chinese_chars,
        mark_sentence_endings,
    ],
    sentseg_start_cond_fn,
    sentseg_end_cond_fn,
)
_sentseg_with_comma = generate_segmenter(
    [
        mark_whitespaces,
        mark_extended_chinese_chars,
        mark_sentence_endings_with_comma,
    ],
    sentseg_start_cond_fn,
    sentseg_end_cond_fn,
)


def sentseg(text, enable_comma=False):
    if enable_comma:
        return _sentseg_with_comma(text)
    else:
        return _sentseg(text)
